#include <GL/glew.h>

#include <GL/glut.h>

#include <math.h>

#pragma comment (lib, "glew32.lib")

typedef unsigned int uint;

/*
Controls:
ESC: Quit.
d: View proxy shapes on top of rendering (debug).

Arrow Keys: Move camera

Mouse Left Drag: Rotate camera

Mouse Right Drag: Zoom in and out
*/

//three parameters to fool around with:

#define light_cutoff 0.01 //light intensity cutoff point
#define NUM_LIGHTS 400 //number of lights
#define LIGHT_BRIGTNESS 20 //brightness of lights

#include "Shader.h"

struct vec3 {
	double x, y, z;
	vec3(double tx=0, double ty=0, double tz=0) : x(tx), y(ty), z(tz) {}
	double length() const {return sqrt(x*x+y*y+z*z);}
	vec3 operator *(double f) const {return vec3(x*f, y*f, z*f);}
};

#include "Camera.h"

enum {POSITION_UNIT, NORMAL_UNIT, COLOR_UNIT, ATTR_UNIT};
enum {S_DEF, S_LIGHT, S_LIGHT_DEBUG, NUM_SHADERS};

int LIGHT_SHADER = S_LIGHT;

uint width = 600, height = 400;

GLuint positionMap, normalMap, colorMap, attrMap, depthMap;

GLuint deferredFBO;

double fov = 45, zNear = 0.1, zFar = 10000;

Shader* shaders[NUM_SHADERS];

Camera camera;

class Light {
private:
	double dist;
	vec3 pos, col;
public:
	Light(const vec3& p, const vec3& c) : pos(p), col(c) {dist = sqrt(col.length()/light_cutoff);}
	void setPos(const vec3& p){pos = p;}
	inline const vec3& getPos() const {return pos;}
	void setCol(const vec3& c){col = c; dist = sqrt(col.length()/light_cutoff);}
	inline const vec3& getCol() const {return col;}
	inline double getDist() const {return dist;}
	void render() const {
		glLoadIdentity();
		glTranslated(pos.x, pos.y, pos.z);
		glScaled(dist, dist, dist);
		glutSolidCube(2);
		glLoadIdentity();
	}
};

Light* lights[NUM_LIGHTS];

GLuint emptyTexture(uint width, uint height, uint iformat, uint format, uint type = GL_UNSIGNED_BYTE);

bool FBO_SAME_FORMAT = false;

void createDeferredMaps(){
	if(FBO_SAME_FORMAT){
		positionMap = emptyTexture(width, height, GL_RGBA_FLOAT32_ATI, GL_RGBA, GL_FLOAT);
		normalMap = emptyTexture(width, height, GL_RGBA_FLOAT32_ATI, GL_RGBA, GL_FLOAT);
		colorMap = emptyTexture(width, height, GL_RGBA_FLOAT32_ATI, GL_RGBA);
		attrMap = emptyTexture(width, height, GL_RGBA_FLOAT32_ATI, GL_RGBA, GL_FLOAT);
		depthMap = emptyTexture(width, height, GL_DEPTH_COMPONENT, GL_DEPTH_COMPONENT);
	} else {
		positionMap = emptyTexture(width, height, GL_RGB_FLOAT32_ATI, GL_RGB, GL_FLOAT);
		normalMap = emptyTexture(width, height, GL_RGB_FLOAT32_ATI, GL_RGB, GL_FLOAT);
		colorMap = emptyTexture(width, height, GL_RGBA, GL_RGBA);
		attrMap = emptyTexture(width, height, GL_RGB_FLOAT32_ATI, GL_RGB, GL_FLOAT);
		depthMap = emptyTexture(width, height, GL_DEPTH_COMPONENT, GL_DEPTH_COMPONENT);
	}

	glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, deferredFBO);

	glFramebufferTexture2DEXT(GL_FRAMEBUFFER_EXT, GL_COLOR_ATTACHMENT0_EXT, GL_TEXTURE_2D, colorMap, 0);
	glFramebufferTexture2DEXT(GL_FRAMEBUFFER_EXT, GL_COLOR_ATTACHMENT1_EXT, GL_TEXTURE_2D, positionMap, 0);
	glFramebufferTexture2DEXT(GL_FRAMEBUFFER_EXT, GL_COLOR_ATTACHMENT2_EXT, GL_TEXTURE_2D, normalMap, 0);
	glFramebufferTexture2DEXT(GL_FRAMEBUFFER_EXT, GL_COLOR_ATTACHMENT3_EXT, GL_TEXTURE_2D, attrMap, 0);
	glFramebufferTexture2DEXT(GL_FRAMEBUFFER_EXT, GL_DEPTH_ATTACHMENT_EXT, GL_TEXTURE_2D, depthMap, 0);

	GLenum buffers[] = {GL_COLOR_ATTACHMENT0_EXT, GL_COLOR_ATTACHMENT1_EXT, GL_COLOR_ATTACHMENT2_EXT, GL_COLOR_ATTACHMENT3_EXT};
	glDrawBuffers(4, buffers);

	glReadBuffer(GL_NONE);

	GLenum status = glCheckFramebufferStatusEXT(GL_FRAMEBUFFER_EXT);

	glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, 0);

	if(!FBO_SAME_FORMAT){
		if(status==GL_FRAMEBUFFER_INCOMPLETE_FORMATS_EXT){
			FBO_SAME_FORMAT = true;
			glDeleteTextures(1, &positionMap);
			glDeleteTextures(1, &normalMap);
			glDeleteTextures(1, &colorMap);
			glDeleteTextures(1, &attrMap);
			glDeleteTextures(1, &depthMap);

			createDeferredMaps();
			return;
		}
	}

	if(status != GL_FRAMEBUFFER_COMPLETE_EXT){printf("ERROR: Your card does not support framebuffer objects with MRTs.\n"); exit(-1);}
}

//initialize
void initgl()
{
	glewInit();

	glGenFramebuffersEXT(1, &deferredFBO);
	createDeferredMaps();

	shaders[S_DEF] = new Shader("shaders/deferred");
	shaders[S_LIGHT] = new Shader("shaders/light");
	shaders[S_LIGHT_DEBUG] = new Shader("shaders/light_debug");

	glClearDepth(1.0f);
	glDisable(GL_CULL_FACE);
	glEnable(GL_DEPTH_TEST);
	glDepthFunc(GL_LEQUAL);
	glShadeModel(GL_SMOOTH);
	glDisable(GL_LIGHTING);
	glClearColor(0, 0, 0, 0);
	glColor4d(1, 1, 1, 1);
	glHint(GL_PERSPECTIVE_CORRECTION_HINT, GL_NICEST);
	glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);

	int ln = 0;
	for(double y = 100; y<=3100; y+=2900.0/(NUM_LIGHTS/2))
		for(int x = 1490; x<=1510; x+= 20)
			if(ln < NUM_LIGHTS)
				lights[ln++] = new Light(vec3(x, y, 2.5), vec3(1, 0.3, 0.086)*LIGHT_BRIGTNESS);
}

GLuint emptyTexture(uint width, uint height, uint iformat, uint format, uint type){
	GLuint ret;
	glGenTextures(1, &ret);
	glBindTexture(GL_TEXTURE_2D, ret);
	glTexImage2D(GL_TEXTURE_2D, 0, iformat, width, height, 0, format, type, NULL);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP);
	return ret;
}

void bindTexture(uint tex, uint unit, uint target = GL_TEXTURE_2D){
	glActiveTexture(GL_TEXTURE0 + unit);
	glBindTexture(target, tex);
}

void unbindTexture(uint unit, uint target = GL_TEXTURE_2D){
	glActiveTexture(GL_TEXTURE0 + unit);
	glBindTexture(target, 0);
}

inline void setVec(Shader* s, const char* l, const vec3& v){s->set3f(l, (float)v.x, (float)v.y, (float)v.z);}

void genDeferred(){
	glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, deferredFBO);
	glViewport(0, 0, width, height);
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
	Shader* s = shaders[S_DEF];

	s->bind();
	s->setf("diff_coeff", 1);
	s->setf("phong_coeff", 0);
	s->setf("two_sided", 0.8f);

	for(int i=0; i<NUM_LIGHTS; i++){
		glLoadIdentity();
		vec3 pos = lights[i]->getPos();
		glTranslated(pos.x, pos.y, pos.z);
		glScaled(2, 2, 2.5);
		glutSolidCube(2);
		glLoadIdentity();
	}

	s->setf("two_sided", 0);

	glBegin(GL_QUADS);
	glNormal3f(0, 0, 1);
	glVertex3f(0, 0, 0);
	glVertex3f(3000, 0, 0);
	glVertex3f(3000, 3000, 0);
	glVertex3f(0, 3000, 0);
	glEnd();

	s->unbind();

	glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, 0);
}

Shader* cur_shader;

void renderLight(Light* l){
	setVec(cur_shader, "lightPos", l->getPos()); //pass light pos to pixel shader
	setVec(cur_shader, "lightCol", l->getCol()); //pass light color (magnitude is power) to pixel shader
	cur_shader->setf("dist", l->getDist()); //pass the light's area of effect radius to pixel shader
	l->render(); //render the proxy shape
}
void renderLights(){
	glEnable(GL_BLEND);
	glBlendFunc(GL_ONE, GL_ONE); //set blend function
	glEnable(GL_CULL_FACE); //cull front faces
	glCullFace(GL_FRONT);
	glDisable(GL_DEPTH_TEST); //disable depth testing
	glDepthMask(false);
	Shader* s = shaders[LIGHT_SHADER]; //bind point light pixel shader
	s->bind();
	bindTexture(positionMap, POSITION_UNIT); //bind position, normal and color textures from deferred shading pass
	s->seti("positionMap", POSITION_UNIT);
	bindTexture(normalMap, NORMAL_UNIT);
	s->seti("normalMap", NORMAL_UNIT);
	bindTexture(colorMap, COLOR_UNIT);
	s->seti("colorMap", COLOR_UNIT);
	bindTexture(attrMap, ATTR_UNIT);
	s->seti("attrMap", ATTR_UNIT);
	camera.loadCameraPos(s, "camPos"); //set pixel shader attribute "camPos" to the camera pos (as it is needed for phong shading)
	cur_shader = s;
	for(int i=0; i<NUM_LIGHTS; i++)
		renderLight(lights[i]);
	s->unbind(); //unbind and reset everything to desired values
	unbindTexture(POSITION_UNIT);
	unbindTexture(NORMAL_UNIT);
	unbindTexture(COLOR_UNIT);
	unbindTexture(ATTR_UNIT);
	glDisable(GL_CULL_FACE);
	glEnable(GL_DEPTH_TEST);
	glDepthMask(true);
	glDisable(GL_BLEND);
}

//display frame
void display()
{
	glClear(GL_COLOR_BUFFER_BIT);

	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	gluPerspective(fov, ((double)width)/height, zNear, zFar);
	camera.loadCamera();

	glMatrixMode(GL_MODELVIEW);

	glLoadIdentity();

	//generate deferred shading maps
	genDeferred();

	//render scene normally with whatever effects and light sources you want
	//this demo only renders point light sources, so nothing happens here.

	//add the contribution of the point light sources.
	renderLights();

	glFlush();

	glutSwapBuffers();
}

//reshape window
void reshape( int w, int h )
{
	if(w == width && height == h) return;
	width = w, height = h;

	glViewport(0, 0, width, height);

	glDeleteTextures(1, &positionMap);
	glDeleteTextures(1, &normalMap);
	glDeleteTextures(1, &colorMap);
	glDeleteTextures(1, &attrMap);
	glDeleteTextures(1, &depthMap);

	createDeferredMaps();
	
	glutPostRedisplay();
}

void keyPress( unsigned char key, int x, int y ){
	if(key==27) exit(0);
	else if(key == 'd'){
		LIGHT_SHADER = (LIGHT_SHADER == S_LIGHT)?S_LIGHT_DEBUG:S_LIGHT;
		glutPostRedisplay();
	}
}
void specialKey(int key, int x, int y){camera.keyPress(key, x, y);}
void mouse( int b, int state, int x, int y ){camera.mouse(b, state, x, y);}
void dragged( int x, int y ){camera.dragged(x, y);}

int main (int argc, char *argv[]) {
	
	glutInit(&argc, argv);
	glutInitDisplayMode(GLUT_RGBA | GLUT_DOUBLE | GLUT_DEPTH);
	glutInitWindowSize(width, height);

	glutCreateWindow("Image Space Lighting Demo");
	
	glutDisplayFunc(display);
	glutReshapeFunc(reshape);
	glutKeyboardFunc(keyPress);
	glutSpecialFunc(specialKey);
	glutMouseFunc(mouse);
	glutMotionFunc(dragged);
	glutPassiveMotionFunc(dragged);
	
	initgl();

	glutMainLoop();
}

